@@ -26,7 +26,9 @@
 #include <KDirWatch>
 #include <KSharedConfig>
 
+#include <QAction>
 #include <QModelIndex>
+#include <QMimeData>
 #include <QSet>
 
 using namespace Milou;
@@ -58,6 +60,7 @@ QHash<int, QByteArray> SourcesModel::roleNames() const
     QHash<int, QByteArray> roles = QAbstractListModel::roleNames();
     roles.insert(TypeRole, "type");
     roles.insert(SubtextRole, "subtext");
+    roles.insert(ActionsRole, "actions");
     roles.insert(DuplicateRole, "isDuplicate");
     roles.insert(PreviewTypeRole, "previewType");
     roles.insert(PreviewUrlRole, "previewUrl");
@@ -108,6 +111,21 @@ QVariant SourcesModel::data(const QModelIndex& index, int role) const
         case SubtextRole:
             return m.subtext();
 
+        case ActionsRole: {
+            const auto &actions = m_manager->actionsForMatch(m);
+            if (actions.isEmpty()) {
+                return QVariantList();
+            }
+
+            QVariantList actionsList;
+            actionsList.reserve(actions.size());
+
+            for (QAction *action : actions) {
+                actionsList.append(QVariant::fromValue(action));
+            }
+
+            return actionsList;
+        }
         case DuplicateRole:
             return m_duplicates.value(m.text());
 
@@ -151,7 +169,34 @@ QString SourcesModel::runner() const
 
 void SourcesModel::setRunner(const QString& runner)
 {
-    m_runner = runner;
+    if (m_runner != runner) {
+        m_runner = runner;
+
+        m_manager->setSingleModeRunnerId(m_runner);
+        m_manager->setSingleMode(!m_runner.isEmpty());
+
+        emit runnerChanged();
+    }
+}
+
+QString SourcesModel::runnerName() const
+{
+    auto *singleRunner = m_manager->singleModeRunner();
+    if (!singleRunner) {
+        return QString();
+    }
+
+    return singleRunner->name();
+}
+
+QIcon SourcesModel::runnerIcon() const
+{
+    auto *singleRunner = m_manager->singleModeRunner();
+    if (!singleRunner) {
+        return QIcon();
+    }
+
+    return singleRunner->icon();
 }
 
 void SourcesModel::setQueryLimit(int limit)
@@ -176,8 +221,6 @@ void SourcesModel::setQueryString(const QString& str)
     }
 
     m_modelPopulated = false;
-    m_manager->setSingleModeRunnerId(m_runner);
-    m_manager->setSingleMode(!m_runner.isEmpty());
     m_manager->launchQuery(m_queryString, m_runner);
 
     // We avoid clearing the model instantly, and instead wait for the results
@@ -380,8 +423,36 @@ bool SourcesModel::run(int index)
     return true;
 }
 
+bool SourcesModel::runAction(int index, int actionIndex)
+{
+    Plasma::QueryMatch match = fetchMatch(index);
+    Q_ASSERT(match.runner());
+
+    const auto &actions = m_manager->actionsForMatch(match);
+    if (actionIndex < 0 || actionIndex >= actions.count()) {
+        return false;
+    }
+
+    QAction *action = actions.at(actionIndex);
+    match.setSelectedAction(action);
+    m_manager->run(match);
+    return true;
+}
+
 void SourcesModel::reloadConfiguration()
 {
     KSharedConfig::openConfig("krunnerrc")->reparseConfiguration();
     m_manager->reloadConfiguration();
 }
+
+QMimeData *SourcesModel::getMimeData(int index) const
+{
+    Plasma::QueryMatch match = fetchMatch(index);
+    Q_ASSERT(match.runner());
+
+    // we're returning a parent-less QObject from a Q_INVOKABLE
+    // which means the QML engine will take care of deleting it eventually
+    QMimeData *mimeData = m_manager->mimeDataForMatch(match);
+
+    return mimeData;
+}
